/*
Eric Jardim
IMPA - Instituto de Matematica Pura e Aplicada

ar.cc
*/

#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <unistd.h>
#include "TriMesh.h"
#include "TriMesh_algo.h"
#include "XForm.h"
#include "GLCamera.h"
#include "ICP.h"
#include <GL/glew.h>
#include <GL/glut.h>
#include <string>
#include <sstream>
#include <cmath>
#include <deque>
using namespace std;

#include "apparentridge.h"

#define FBO
#define VBO_


// Globals
TriMesh *mesh;
GLuint dlist;
xform xf;
string xffilename;
string base_name, modelname;
TriMesh::BSphere global_bsph;
xform global_xf;
GLCamera camera;

GLfloat *curv;
GLfloat *curvatures;
vector<GLfloat> ndotv;
float tau, radius;
float k1max, k2max, k1min, k2min, k1MAX, k2MAX, k1mean, k2mean, k1var, k2var;
float feature_size;
int test_ar = 1;
GLuint tau_location;
float ar_thresh = 0.125;
int ar_n = 8;
float lambda = 0.1;
int lambda_n = 0;
float kappa = 3;
int kappa_n = 6;
int mode = 0;
int fade = 1;
float line_width = 4.0;

vec currcolor(0, 0, 0); // Current line color
GLhandleARB ar_vs, ar_fs_1p, ar_fs_2p, ar_fs_3p, pass1_program, pass2_program, pass3_program;
GLuint buffers[3] = {GL_BACK_LEFT, GL_AUX0, GL_AUX1};
GLuint depth_rbo, fbo, vertex_vbo, normal_vbo, curv_vbo, pdir_vbo, index_vbo;
int pdir1_location, pdir2_location;
vector<float> q1;
vector<vec2> t1;
vector<float> Dt1q1;

GLboolean npotTexturesAvailable = GL_FALSE;
GLint windowWidth = 640;
GLint windowHeight = 640;
GLint textureWidth = windowWidth;
GLint textureHeight = windowHeight;
GLint maxTexSize;
GLfloat texCoordOffsets[18];
GLuint textures[3];

enum DrawState {NONE, AR, JUDD_AR, LAMBERTIAN, Q1};
enum ShaderType {VERTEX_SHADER, FRAGMENT_SHADER};

bool draw_edges = false;
bool draw_2side = true;
bool draw_shiny = false;
bool draw_lit = false;
bool draw_index = false;
bool draw_smooth = true;
bool boundaries = false;
bool do_image_processing = true;
bool do_post_processing = false;
bool white_bg = false;
DrawState draw_mode = AR, last_draw_mode = NONE;
bool draw_faded = true;


class Chronometer {
    timestamp t;
    deque<float> dt;
    
public:
    void tic() {
        t = now();
    }
    
    void tac() {
        dt.push_back( now()-t );
    }
    double mean()
    {
        static int max_frames = 5000;
        if (dt.empty()) return 0.0;
        if (dt.size() > max_frames) dt.pop_front();
        double secs = 0.0;
        for(int i=0; i<dt.size(); i++) secs += dt[i];
        return secs / (double) dt.size();
        //return (float) dt.size() / secs;
    }
    void clear() {
        dt.clear();
    }
};

Chronometer cm, cm1, cm2;

void clear_chronometers() {
    cm.clear();
    cm1.clear();
    cm2.clear();
}

void render_list();

// Signal a redraw
void need_redraw()
{
    glutPostRedisplay();
}


// Get the base name from the model filename
static inline std::string
getbase_name(const std::string &filename)
{
    std::string x = filename;
    std::string::size_type dot = x.rfind(".", x.length());
    if (dot != std::string::npos)
        x.erase(dot);
    return x;
}

// Get the base name from the model filename
static inline std::string
get_model_name(const std::string &filename)
{
    std::string x = filename;
    std::string::size_type bar = x.rfind("/", x.length());
    if (bar != std::string::npos)
        x.erase(0,bar+1);
        //x.erase(bar);
    return x;
}

// Generate a filename for saving the current state
// filename := basepath<suffix><slot>
static inline std::string
savename(const std::string &base, const std::string &suffix, int slot)
{
    stringstream ss;
    ss << base << suffix;
    ss.width(2);
    ss.fill('0');
    ss << slot;
    return ss.str();
}


void saveparams(const std::string &filename)
{
    std::ofstream f(filename.c_str());
    f << tau << std::endl << lambda << std::endl;
    f.close();
}

void loadparams(const std::string &filename)
{
    std::ifstream f(filename.c_str());
    float t, l;
    f >> t;
    f >> l;
    f.close();
    if (f.good()) {
        //tau = t;
        //lambda = l;
        printf("tau=%0.2f, lambda=%0.2f\n", tau, lambda);
    }
}


// Clear the screen
void cls()
{
    glDisable(GL_DITHER);
    glDisable(GL_BLEND);
    glDisable(GL_DEPTH_TEST);
    glDisable(GL_NORMALIZE);
    glDisable(GL_LIGHTING);
    glDisable(GL_NORMALIZE);
    glDisable(GL_COLOR_MATERIAL);
    if (white_bg || draw_mode == JUDD_AR || draw_mode == Q1)// || draw_mode == AR)
        glClearColor(1, 1, 1, 0);
    else
        //glClearColor(0.08, 0.08, 0.08, 0);
        glClearColor(-100.0, 0.0, 0.0, 0.0);
    glClearDepth(1.0f);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
}


// Set up lights and materials
void setup_lighting()
{
    Color c(1.0f);
    /*if (draw_falsecolor)
        c = Color::hsv(-3.88, 0.6, 1.0);*/
    glColor3fv(c);

    if (!draw_lit || mesh->normals.empty()) {
        glDisable(GL_LIGHTING);
        return;
    }

    GLfloat mat_specular[4] = { 0.18, 0.18, 0.18, 0.18 };
    if (!draw_shiny) {
        mat_specular[0] = mat_specular[1] =
        mat_specular[2] = mat_specular[3] = 0.0f;
    }
    GLfloat mat_shininess[] = { 64 };
    GLfloat global_ambient[] = { 0.02, 0.02, 0.05, 0.05 };
    GLfloat light0_ambient[] = { 0, 0, 0, 0 };
    GLfloat light0_diffuse[] = { 0.85, 0.85, 0.8, 0.85 };
    GLfloat light1_diffuse[] = { -0.01, -0.01, -0.03, -0.03 };
    GLfloat light0_specular[] = { 0.85, 0.85, 0.85, 0.85 };
    glColorMaterial(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE);
    glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, mat_specular);
    glMaterialfv(GL_FRONT_AND_BACK, GL_SHININESS, mat_shininess);
    glLightfv(GL_LIGHT0, GL_AMBIENT, light0_ambient);
    glLightfv(GL_LIGHT0, GL_DIFFUSE, light0_diffuse);
    glLightfv(GL_LIGHT0, GL_SPECULAR, light0_specular);
    glLightfv(GL_LIGHT1, GL_DIFFUSE, light1_diffuse);
    glLightModelfv(GL_LIGHT_MODEL_AMBIENT, global_ambient);
    glLightModeli(GL_LIGHT_MODEL_LOCAL_VIEWER, GL_FALSE);
    glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, draw_2side);
    glEnable(GL_LIGHTING);
    glEnable(GL_LIGHT0);
    glEnable(GL_LIGHT1);
    glEnable(GL_COLOR_MATERIAL);
    glEnable(GL_NORMALIZE);
}


void image_processing()
{
    // Redraw from texture w/ fragment shader
    GLint uniformLoc;
    glUseProgramObjectARB(pass2_program);
    uniformLoc = glGetUniformLocation(pass2_program, "sampler0");
    if (uniformLoc != -1) {
        glUniform1i(uniformLoc, 0);
    }
    uniformLoc = glGetUniformLocation(pass2_program, "tc_offset");
    if (uniformLoc != -1) {
        glUniform2fv(uniformLoc, 9, texCoordOffsets);
    }
    uniformLoc = glGetUniformLocation(pass2_program, "lambda");
    if (uniformLoc != -1) {
        glUniform1fv(uniformLoc, 1, &lambda);
    }
    uniformLoc = glGetUniformLocation(pass2_program, "tau");
    if (uniformLoc != -1) {
        glUniform1fv(uniformLoc, 1, &tau);
    }
    /*uniformLoc = glGetUniformLocation(pass2_program, "kappa");
    if (uniformLoc != -1) {
        glUniform1fv(uniformLoc, 1, &kappa);
    }*/
    uniformLoc = glGetUniformLocation(pass2_program, "mode");
    if (uniformLoc != -1) {
        glUniform1iv(uniformLoc, 1, &mode);
    }
    uniformLoc = glGetUniformLocation(pass2_program, "fade");
    if (uniformLoc != -1) {
        glUniform1iv(uniformLoc, 1, &fade);
    }

    glDisable(GL_CULL_FACE);
    glDisable(GL_DEPTH_TEST);
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

    glBindTexture(GL_TEXTURE_2D, textures[0]);
    #ifndef FBO
    glActiveTexture(GL_TEXTURE0);
    //glReadBuffer(GL_BACK_LEFT); 
    glCopyTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, (windowWidth-textureWidth)/2,
                         (windowHeight-textureHeight)/2, textureWidth, textureHeight, 0);
    #endif
                         
    /*glActiveTexture(GL_TEXTURE1);
    glReadBuffer(GL_AUX0); 
    glBindTexture(GL_TEXTURE_2D, textures[1]);
    glCopyTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, (windowWidth-textureWidth)/2,
                         (windowHeight-textureHeight)/2, textureWidth, textureHeight, 0);

    glActiveTexture(GL_TEXTURE2);
    glReadBuffer(GL_AUX1); 
    glBindTexture(GL_TEXTURE_2D, textures[2]);
    glCopyTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, (windowWidth-textureWidth)/2,
                         (windowHeight-textureHeight)/2, textureWidth, textureHeight, 0);*/
    
    //glActiveTexture(GL_TEXTURE0);
   
    //glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    //glClear(GL_DEPTH_BUFFER_BIT);

    //glBindTexture(GL_TEXTURE_2D, textures[0]);
    //glBindTexture(GL_TEXTURE_2D, 0);
    
    //cls();
    //glClearDepth(1.0f);
    //glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    
    glBegin(GL_QUADS);
    //
    /*glMultiTexCoord2f(GL_TEXTURE0, 0.0f, 0.0f);
    glVertex2f(-1.0, -1.0);
    glMultiTexCoord2f(GL_TEXTURE0, 0.0f, 1.0f);
    glVertex2f(-1.0, 1.0);
    glMultiTexCoord2f(GL_TEXTURE0, 1.0f, 1.0f);
    glVertex2f(1.0, 1.0);
    glMultiTexCoord2f(GL_TEXTURE0, 1.0f, 0.0f);
    glVertex2f(1.0, -1.0);*/
    //
    glMultiTexCoord2f(GL_TEXTURE0, 0.0f, 0.0f);
    glVertex2f(-((GLfloat)textureWidth / (GLfloat)windowWidth), -((GLfloat)textureHeight / (GLfloat)windowHeight));
    glMultiTexCoord2f(GL_TEXTURE0, 0.0f, 1.0f);
    glVertex2f(-((GLfloat)textureWidth / (GLfloat)windowWidth), ((GLfloat)textureHeight / (GLfloat)windowHeight));
    glMultiTexCoord2f(GL_TEXTURE0, 1.0f, 1.0f);
    glVertex2f(((GLfloat)textureWidth / (GLfloat)windowWidth), ((GLfloat)textureHeight / (GLfloat)windowHeight));
    glMultiTexCoord2f(GL_TEXTURE0, 1.0f, 0.0f);
    glVertex2f(((GLfloat)textureWidth / (GLfloat)windowWidth), -((GLfloat)textureHeight / (GLfloat)windowHeight));
    //
    glEnd();
    // <-
    
    glUseProgramObjectARB(0);
    glEnable(GL_DEPTH_TEST);
}

void just_show() 
{
    glDisable(GL_CULL_FACE);
    glDisable(GL_DEPTH_TEST);
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

    glBindTexture(GL_TEXTURE_2D, textures[0]);  
    glActiveTexture(GL_TEXTURE0);

    glBegin(GL_QUADS);
    glMultiTexCoord2f(GL_TEXTURE0, 0.0f, 0.0f);
    glVertex2f(-1.0, -1.0);
    glMultiTexCoord2f(GL_TEXTURE0, 0.0f, 1.0f);
    glVertex2f(-1.0, 1.0);
    glMultiTexCoord2f(GL_TEXTURE0, 1.0f, 1.0f);
    glVertex2f(1.0, 1.0);
    glMultiTexCoord2f(GL_TEXTURE0, 1.0f, 0.0f);
    glVertex2f(1.0, -1.0);
    glEnd();
    
    glEnable(GL_DEPTH_TEST);
}

void post_processing()
{
    GLint uniformLoc;
    glUseProgramObjectARB(pass3_program);
    uniformLoc = glGetUniformLocation(pass2_program, "sampler0");
    if (uniformLoc != -1) {
        glUniform1i(uniformLoc, 0);
    }
    uniformLoc = glGetUniformLocation(pass2_program, "tc_offset");
    if (uniformLoc != -1) {
        glUniform2fv(uniformLoc, 9, texCoordOffsets);
    }
    uniformLoc = glGetUniformLocation(pass2_program, "lambda");
    if (uniformLoc != -1) {
        glUniform1fv(uniformLoc, 1, &lambda);
    }
    uniformLoc = glGetUniformLocation(pass2_program, "tau");
    if (uniformLoc != -1) {
        glUniform1fv(uniformLoc, 1, &tau);
    }

    glDisable(GL_CULL_FACE);
    glDisable(GL_DEPTH_TEST);
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();


    

    glBindTexture(GL_TEXTURE_2D, textures[0]);
    glActiveTexture(GL_TEXTURE0);
    //glBindTexture(GL_TEXTURE_2D, 0);  
    //glReadBuffer(GL_BACK_LEFT); 
    //glCopyTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, (windowWidth-textureWidth)/2,
    //                     (windowHeight-textureHeight)/2, textureWidth, textureHeight, 0);

    glBegin(GL_QUADS);
    glMultiTexCoord2f(GL_TEXTURE0, 0.0f, 0.0f);
    glVertex2f(-1.0, -1.0);
    glMultiTexCoord2f(GL_TEXTURE0, 0.0f, 1.0f);
    glVertex2f(-1.0, 1.0);
    glMultiTexCoord2f(GL_TEXTURE0, 1.0f, 1.0f);
    glVertex2f(1.0, 1.0);
    glMultiTexCoord2f(GL_TEXTURE0, 1.0f, 0.0f);
    glVertex2f(1.0, -1.0);
    glEnd();
    
    glUseProgramObjectARB(0);
    glEnable(GL_DEPTH_TEST);
}


// Draw triangle strips.  They are stored as length followed by values.
void draw_tstrips(const TriMesh *themesh)
{
    #ifdef VBO
    glBindBufferARB(GL_ELEMENT_ARRAY_BUFFER_ARB, index_vbo);
    const int *t = &themesh->tstrips[0];
    const int *start = t;
    const int *end = t + themesh->tstrips.size();
    while (likely(t < end)) {
        int striplen = *t++;
        glDrawElements(GL_TRIANGLE_STRIP, striplen, GL_UNSIGNED_INT, (GLvoid*)(t-start));
        t += striplen;
    }
    glBindBufferARB(GL_ELEMENT_ARRAY_BUFFER_ARB, 0);
    #else
    const int *t = &themesh->tstrips[0];
    const int *end = t + themesh->tstrips.size();
    while (likely(t < end)) {
        int striplen = *t++;
        glDrawElements(GL_TRIANGLE_STRIP, striplen, GL_UNSIGNED_INT, t);
        t += striplen;
    }
    #endif
}

// Prepare the display
void prepare_list() {
    glNewList(dlist, GL_COMPILE);
    render_list();
    glEndList();
}


// Draw the boundaries on the mesh
void draw_boundaries()
{
    mesh->need_faces();
    mesh->need_across_edge();

    if (draw_mode == AR) 
        glColor4f(0.0, 0.0, 0.0, 1.0);
    else 
        glColor3f(0.0, 0.0, 0.0);
    glLineWidth(0.666*line_width);

    glBegin(GL_LINES);
    for (int i = 0; i < mesh->faces.size(); i++) {
        for (int j = 0; j < 3; j++) {
            if (mesh->across_edge[i][j] >= 0)
                continue;
            int v1 = mesh->faces[i][(j+1)%3];
            int v2 = mesh->faces[i][(j+2)%3];
            glVertex3fv(mesh->vertices[v1]);
            glVertex3fv(mesh->vertices[v2]);
        }
    }
    glEnd();
}

// Draw the mesh
void draw_mesh()
{
    glPushMatrix();
    glMultMatrixd(xf);
    glDepthFunc(GL_LESS);
    glEnable(GL_DEPTH_TEST);

    if (draw_2side) {
        glDisable(GL_CULL_FACE);
    } 
    else {
        glCullFace(GL_BACK);
        glEnable(GL_CULL_FACE);
    }

    // Vertices
    #ifdef VBO
    glBindBufferARB(GL_ARRAY_BUFFER_ARB, vertex_vbo);
    glEnableClientState(GL_VERTEX_ARRAY);
    glVertexPointer(3, GL_FLOAT,0,0);
    #else
    glEnableClientState(GL_VERTEX_ARRAY);
    glVertexPointer(3, GL_FLOAT,sizeof(mesh->vertices[0]),&mesh->vertices[0][0]);
    #endif
  

    // Normals
    if (!mesh->normals.empty() && !draw_index) {
        #ifdef VBO        
        glBindBufferARB(GL_ARRAY_BUFFER_ARB, normal_vbo);
        glEnableClientState(GL_NORMAL_ARRAY);
        glNormalPointer(GL_FLOAT, 0, 0);
        #else
        glEnableClientState(GL_NORMAL_ARRAY);
        glNormalPointer(GL_FLOAT, sizeof(mesh->normals[0]), &mesh->normals[0][0]);
        #endif
    } 
    else {
        glDisableClientState(GL_NORMAL_ARRAY);
    }

    if (draw_mode == AR) {
        #ifdef VBO        
        glBindBufferARB(GL_ARRAY_BUFFER_ARB, curv_vbo);
        glEnableClientState(GL_COLOR_ARRAY);
        glColorPointer(3, GL_FLOAT, 0, 0);

        glBindBufferARB(GL_ARRAY_BUFFER_ARB, pdir_vbo);
        glEnableClientState(GL_SECONDARY_COLOR_ARRAY);
        glSecondaryColorPointer(3, GL_FLOAT, 0, 0);
        #else
        glEnableClientState(GL_COLOR_ARRAY);
        glColorPointer(3, GL_FLOAT, sizeof(GLfloat)*3, curvatures);

        glEnableClientState(GL_SECONDARY_COLOR_ARRAY);
        glSecondaryColorPointer(3, GL_FLOAT, sizeof(mesh->pdir1[0]), &mesh->pdir1[0][0]);
        #endif
        
        //glEnableClientState(GL_TEXTURE_COORD_ARRAY);
        //glActiveTexture(GL_TEXTURE2);
        //glClientActiveTexture(GL_TEXTURE2);
        //glTexCoordPointer(3, GL_FLOAT, sizeof(mesh->pdir2[0]), &mesh->pdir2[0][0]);
    }    
    else {
        glDisableClientState(GL_COLOR_ARRAY);
        glDisableClientState(GL_SECONDARY_COLOR_ARRAY);
        glDisableClientState(GL_TEXTURE_COORD_ARRAY);
    }

    // Lambertian shading
    if (draw_mode == LAMBERTIAN) {
        glEnableClientState(GL_COLOR_ARRAY);
        glColorPointer(3, GL_FLOAT,
                sizeof(GLfloat)*3,
                curv);
        glDisable(GL_CULL_FACE);
        glPolygonMode(GL_BACK, GL_FILL);
    }

    // Colors
    //if (!themesh->colors.empty() && !draw_falsecolor && !draw_index) {
    if (draw_mode == Q1) {
        glEnableClientState(GL_COLOR_ARRAY);
        glColorPointer(3, GL_FLOAT,
                sizeof(GLfloat)*3,
                curv);
    }

    // smooth lines
    if (draw_smooth && (draw_mode != AR)) {
        glEnable(GL_LINE_SMOOTH);
        glEnable(GL_BLEND);
        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    }

    // Main drawing pass
    if (draw_edges) {
        glPolygonOffset(10.0f, 10.0f);
        glEnable(GL_POLYGON_OFFSET_FILL);
    }

    
    #ifdef VBO        
    glBindBufferARB(GL_ARRAY_BUFFER_ARB, 0);
    #endif
    if (draw_mode == AR) {
        //glDrawBuffers(3, buffers);
        glCallList(dlist);
        //draw_tstrips(mesh);
    }
    else
        draw_tstrips(mesh);


    if (draw_mode == JUDD_AR) {
        glLineWidth(line_width);
        glBegin(GL_LINES);
        draw_mesh_app_ridges(ndotv, q1, t1, Dt1q1, true, test_ar, lambda / sqr(feature_size));
        glEnd();
        if (boundaries) draw_boundaries(); // <----
    }

    //timestamp before = now();
    if (draw_mode == AR) {
        glUseProgramObjectARB(0);
        if (boundaries) draw_boundaries(); // <----
    }
    //fps.push_back((now() - before));

    glDisable(GL_POLYGON_OFFSET_FILL);

    // Edge drawing pass
    glLineWidth(1.0);
    if (draw_edges) {
        glPolygonMode(GL_FRONT, GL_LINE);
        glDisableClientState(GL_COLOR_ARRAY);
        glDisable(GL_COLOR_MATERIAL);
        GLfloat global_ambient[] = { 0.2, 0.2, 0.2, 1.0 };
        GLfloat light0_diffuse[] = { 0.8, 0.8, 0.8, 0.0 };
        GLfloat light1_diffuse[] = { -0.2, -0.2, -0.2, 0.0 };
        GLfloat light0_specular[] = { 0.0f, 0.0f, 0.0f, 0.0f };
        glLightModelfv(GL_LIGHT_MODEL_AMBIENT, global_ambient);
        glLightfv(GL_LIGHT0, GL_DIFFUSE, light0_diffuse);
        glLightfv(GL_LIGHT1, GL_DIFFUSE, light1_diffuse);
        glLightfv(GL_LIGHT0, GL_SPECULAR, light0_specular);
        GLfloat mat_diffuse[4] = { 0.0f, 0.0f, 1.0f, 1.0f };
        glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, mat_diffuse);
        glColor3f(0, 0, 1); // Used iff unlit
        draw_tstrips(mesh);
        glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
    }

    glPopMatrix();
}


void render_list() {
    //glDepthFunc(GL_LESS);
    //glEnable(GL_DEPTH_TEST);
    //glDisable(GL_CULL_FACE);

    // vertices
    #ifdef VBO
    glBindBufferARB(GL_ARRAY_BUFFER_ARB, vertex_vbo);
    glEnableClientState(GL_VERTEX_ARRAY);
    glVertexPointer(3, GL_FLOAT, 0, 0);
    #else
    glEnableClientState(GL_VERTEX_ARRAY);
    glVertexPointer(3, GL_FLOAT, sizeof(mesh->vertices[0]), &mesh->vertices[0][0]);
    #endif

    // normals
    #ifdef VBO
    glBindBufferARB(GL_ARRAY_BUFFER_ARB, normal_vbo);
    glEnableClientState(GL_NORMAL_ARRAY);
    glNormalPointer(GL_FLOAT, 0, 0);
    #else
    glEnableClientState(GL_NORMAL_ARRAY);
    glNormalPointer(GL_FLOAT, sizeof(mesh->normals[0]), &mesh->normals[0][0]);
    #endif

    // reset buffer array
    #ifdef VBO
    glBindBufferARB(GL_ARRAY_BUFFER_ARB, 0);
    #endif

    // k1 & k2
    glEnableClientState(GL_COLOR_ARRAY);
    glColorPointer(3, GL_FLOAT, sizeof(GLfloat)*3, curvatures);

    // 1st principal direction (pdir1)
    glEnableClientState(GL_SECONDARY_COLOR_ARRAY);
    glSecondaryColorPointer(3, GL_FLOAT, sizeof(mesh->pdir1[0]), &mesh->pdir1[0][0]);

    //glEnableClientState(GL_TEXTURE_COORD_ARRAY);
    //glActiveTexture(GL_TEXTURE2);
    //glClientActiveTexture(GL_TEXTURE0);
    //glTexCoordPointer(3, GL_FLOAT, sizeof(mesh->pdir2[0]), &mesh->pdir2[0][0]);

    draw_tstrips(mesh);
    //glBindBufferARB(GL_ARRAY_BUFFER_ARB, 0);

    //timestamp before = now();
    //glUseProgramObjectARB(0);
    //if (boundaries) draw_boundaries(); // <----

    //glDisable(GL_POLYGON_OFFSET_FILL);
}



/*float framerate()
{
    static int max_frames = 1000;
    if (fps.empty())
        return 0.0;
    if (fps.size() > max_frames)
        fps.pop_front();
    float secs = 0.0;
    for(int i=0; i<fps.size(); i++)
        secs += fps[i];
    //return 1000.0f * secs / (float) fps.size();
    return (float) fps.size() / secs;
}*/


void compute_perview()
{
    point viewpos = inv(global_xf) * point(0,0,0);
    int nv = mesh->vertices.size();

    #pragma omp parallel for
    for (int i=0; i<nv; i++) {
        vec viewdir = viewpos - mesh->vertices[i];
        float rlv = 1.0f / len(viewdir);
        viewdir *= rlv;
        ndotv[i] = viewdir DOT mesh->normals[i];

        // coordinate system on tangent plane
        float u = viewdir DOT mesh->pdir1[i], u2 = u*u;
        float v = viewdir DOT mesh->pdir2[i], v2 = v*v;
        float csc2theta = 1.0f / (u2 + v2);

        compute_viewdep_curv(mesh, i, ndotv[i],
                u2*csc2theta, u*v*csc2theta, v2*csc2theta,
                q1[i], t1[i]);

        if (draw_mode == JUDD_AR || draw_mode == Q1) {
            float val = 1.0 - tau*q1[i];
            curv[3*i] = curv[3*i+1] = curv[3*i+2] = val;
        }

        //if (draw_mode == JUDD_AR) {
        //    compute_Dt1q1(mesh, i, ndotv[i], q1, t1, Dt1q1[i]);
        //}
        
        if (draw_mode == LAMBERTIAN) {
            curv[3*i] = curv[3*i+1] = curv[3*i+2] = ndotv[i];
        }
    }
    #pragma omp parallel for
	for (int i = 0; i < nv; i++) {
		compute_Dt1q1(mesh, i, ndotv[i], q1, t1, Dt1q1[i]);
	}
}


// Draw the scene
void redraw()
{
    cm.tic();

    if (draw_mode == AR) {
        glUseProgramObjectARB(pass1_program);
        tau_location = glGetUniformLocation(pass1_program, "fs");
        glUniform1f(tau_location, feature_size);
    }
    else {
        glUseProgramObjectARB(0);
    }
    camera.setupGL(global_xf * global_bsph.center, global_bsph.r);
    glPushMatrix();
    glMultMatrixd(global_xf);

    #ifdef FBO
    if (draw_mode == AR && do_image_processing) {
        glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, fbo);
        //glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT, 
        //                     GL_TEXTURE_2D, textures[0], 0);
    }
    #endif

    cls();
    if (draw_mode == JUDD_AR) setup_lighting();
    
    if (draw_mode != AR) {
        cm1.tic();
        compute_perview();
        cm1.tac();
    }

    cm2.tic();
    draw_mesh(); // <---
    cm2.tac();

    if (draw_mode == AR) cm1.tic();
    #ifdef FBO
    if (draw_mode == AR && do_image_processing) {
        if (do_post_processing) {
            image_processing();
            glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);
            post_processing();
        }
        else {
            glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);
            image_processing();
        }
    }
    #else
    if (draw_mode == AR && do_image_processing) image_processing();
    #endif
    //if (draw_mode == AR) cm1.tac();

    glPopMatrix();
    /*if (draw_mode == AR) {
        camera.setupGL(global_xf * global_bsph.center, global_bsph.r);
        glMultMatrixd(global_xf);
        //glUseProgramObjectARB(0);
        glLineWidth(line_width);
        if (boundaries) draw_boundaries(); // <----
    }*/


    glutSwapBuffers();
    
    if (draw_mode == AR && do_image_processing)
        camera.setupGL(global_xf * global_bsph.center, global_bsph.r);

    if (draw_mode == AR) cm1.tac();
    cm.tac();
    
    //printf("\r                       \r%.3f fps.", 1000.0*cm.mean());

    printf("\r                                                             \r");
    double t1=0, t2=0, tt = cm.mean();
    printf("%s ", modelname.c_str());
    if (draw_mode == JUDD_AR) {
        t1 = cm1.mean();
        t2 = cm2.mean();
        printf("TJ ");
    }
    else if (draw_mode == AR) {
        t1 = cm2.mean();
        t2 = cm1.mean();
        printf("EJ ");
    }
    printf("T1:%.1f msec, T2:%.1f msec, Total:%.1f msec, FR:%.1f", 
        1000.0*t1, 1000.0*t2, 1000.0*tt, 1.0/tt );

    fflush(stdout);
}


// Update global bounding sphere.
void update_bsph()
{
    point boxmin(1e38, 1e38, 1e38);
    point boxmax(-1e38, -1e38, -1e38);
    //for (int i = 0; i < meshes.size(); i++) {
        point c = xf * mesh->bsphere.center;
        float r = mesh->bsphere.r;
        for (int j = 0; j < 3; j++) {
            boxmin[j] = min(boxmin[j], c[j]-r);
            boxmax[j] = max(boxmax[j], c[j]+r);
        }
    //}
    point &gc = global_bsph.center;
    float &gr = global_bsph.r;
    gc = 0.5f * (boxmin + boxmax);
    gr = 0.0f;
    //for (int i = 0; i < meshes.size(); i++) {
        /*point*/ c = xf * mesh->bsphere.center;
        /*float*/ r = mesh->bsphere.r;
        gr = max(gr, dist(c, gc) + r);
    //}
}


// Set the view...
void resetview()
{
    camera.stopspin();
    //for (int i = 0; i < meshes.size(); i++)
        if (!xf.read(xffilename))
            xf = xform();
    update_bsph();
    global_xf = xform::trans(0, 0, -5.0f * global_bsph.r) *
            xform::trans(-global_bsph.center);

    // Special case for 1 mesh
    if (xf.read(xffilename)) {
        global_xf = xf;
        xf = xform();
    }
}


// Save the current image to a PPM file.
// Uses the next available filename matching filenamepattern
void dump_image()
{
    // Find first non-used filename
    const char filenamepattern[] = "img%d.ppm";
    int imgnum = 0;
    FILE *f;
    while (1) {
        char filename[1024];
        sprintf(filename, filenamepattern, imgnum++);
        f = fopen(filename, "rb");
        if (!f) {
            f = fopen(filename, "wb");
            printf("\n\nSaving image %s... ", filename);
            fflush(stdout);
            break;
        }
        fclose(f);
    }

    // Read pixels
    GLint V[4];
    glGetIntegerv(GL_VIEWPORT, V);
    GLint width = V[2], height = V[3];
    char *buf = new char[width*height*3];
    glPixelStorei(GL_PACK_ALIGNMENT, 1);
    glReadPixels(V[0], V[1], width, height, GL_RGB, GL_UNSIGNED_BYTE, buf);

    // Flip top-to-bottom
    for (int i = 0; i < height/2; i++) {
        char *row1 = buf + 3 * width * i;
        char *row2 = buf + 3 * width * (height - 1 - i);
        for (int j = 0; j < 3 * width; j++)
            swap(row1[j], row2[j]);
    }

    // Write out file
    fprintf(f, "P6\n#\n%d %d\n255\n", width, height);
    fwrite(buf, width*height*3, 1, f);
    fclose(f);
    delete [] buf;

    printf("Done.\n\n");
}


// Save all transforms
void save_xforms()
{
    printf("Writing %s\n", xffilename.c_str());
    global_xf.write(xffilename);
}


// Handle mouse button and motion events
static unsigned buttonstate = 0;
void doubleclick(int button, int x, int y)
{
}


// ***
void mousemotionfunc(int x, int y)
{
    static const Mouse::button physical_to_logical_map[] = {
        Mouse::NONE, Mouse::ROTATE, Mouse::MOVEXY, Mouse::MOVEZ,
        Mouse::MOVEZ, Mouse::MOVEXY, Mouse::MOVEXY, Mouse::MOVEXY,
    };

    Mouse::button b = Mouse::NONE;
    if (buttonstate & (1 << 3))
        b = Mouse::WHEELUP;
    else if (buttonstate & (1 << 4))
        b = Mouse::WHEELDOWN;
    else if (buttonstate & (1 << 30))
        b = Mouse::LIGHT;
    else
        b = physical_to_logical_map[buttonstate & 7];

    camera.mouse(x, y, b,
            global_xf * global_bsph.center, global_bsph.r,
            global_xf);
    //update_bsph();  // necessary?
    if (b != Mouse::NONE)
        need_redraw();
}


void mousebuttonfunc(int button, int state, int x, int y)
{
    static timestamp last_click_time;
    static unsigned last_click_buttonstate = 0;
    static float doubleclick_threshold = 0.25f;

    if (glutGetModifiers() & GLUT_ACTIVE_CTRL)
        buttonstate |= (1 << 30);
    else
        buttonstate &= ~(1 << 30);

    if (state == GLUT_DOWN) {
        buttonstate |= (1 << button);
        if (buttonstate == last_click_buttonstate &&
            now() - last_click_time < doubleclick_threshold) {
            doubleclick(button, x, y);
            last_click_buttonstate = 0;
        } else {
            last_click_time = now();
            last_click_buttonstate = buttonstate;
        }
    } else {
        buttonstate &= ~(1 << button);
    }

    mousemotionfunc(x, y);
}


// Idle callback
void idle()
{
    xform tmp_xf = global_xf;
    /*if (current_mesh >= 0)
        tmp_xf = global_xf * xforms[current_mesh];*/

    if (camera.autospin(tmp_xf)) need_redraw();
    else usleep(1000);

    global_xf = tmp_xf;
}


// Keyboard
#define Ctrl (1-'a')
void keyboardfunc(unsigned char key, int x, int y)
{
    switch (key) {
        case '1':
            draw_mode = AR;
            white_bg = false;
            break;
        case '2':
            draw_mode = JUDD_AR;
            white_bg = false;
            break;
        case '3':
            draw_mode = LAMBERTIAN;
            white_bg = true;
            break;
        case '4':
            draw_mode = Q1;
            break;
        case '-':
            tau -= 0.5;
            glUniform1f(tau_location, feature_size);
            printf("tau: %0.2f\n", tau);
            break;
        case '+':
            tau += 0.5;
            glUniform1f(tau_location, feature_size);
            printf("tau: %0.2f\n", tau);
            break;
        case '[':
            lambda_n--;
            lambda = 0.1*pow(1.2, lambda_n);
            printf("\nlambda: %0.7f, lambda_n: %d\n", lambda, lambda_n);
            break;
        case ']':
            lambda_n++;
            lambda = 0.1*pow(1.2, lambda_n);
            printf("\nlambda: %0.7f, lambda_n: %d\n", lambda, lambda_n);
            break;
        case ',':
            line_width -= 0.5;
            printf("\nline width: %0.1f\n", line_width);
            break;
        case '.':
            line_width += 0.5;
            printf("\nline width: %0.1f\n", line_width);
            break;
        case '{':
            if (--ar_n == 0) ar_n = 1;
            ar_thresh = 1.0 / (float) ar_n;
            printf("ar_n: %d\n", ar_n);
            break;
        case '}':
            ar_thresh = 1.0 / (float) ++ar_n;
            printf("ar_n: %d\n", ar_n);
            break;
        case ' ':
            do_image_processing = !do_image_processing; break;
        case 'p':
            do_post_processing = !do_post_processing; break;
        case 'm':
            mode = (mode+1)%2;
            break;
        case 'b':
            boundaries = !boundaries; break;
        case 'o': // Shift-2
            draw_2side = !draw_2side; break;
        case 'e':
            draw_edges = !draw_edges; break;
        case 'f':
            draw_faded = !draw_faded; 
            fade = draw_faded;
            break;
        case 'l':
            draw_lit = !draw_lit; break;
        case 's':
            draw_shiny = !draw_shiny; break;
        case 'w':
            white_bg = !white_bg; break;
        case 'I':
            dump_image(); break;
        case 'a':
            draw_smooth = !draw_smooth; break;
        case Ctrl+'x':
            save_xforms(); break;
        case 'r':
            clear_chronometers(); break;
        case '\033': // Esc
        case Ctrl+'q':
        case 'Q':
        case 'q':
            exit(0);
        default:
            printf("key: %d\n", key);
    }
    if (draw_mode != last_draw_mode) {
        last_draw_mode = draw_mode;
        clear_chronometers();
    }
    need_redraw();
}

void save_view(int slot)
{
    printf("Save viewpoint #%d\n", slot);
//     global_xf.write(vp_filenames[slot-1]);
    global_xf.write(savename(base_name, ".v", slot));
    saveparams(savename(base_name, ".p", slot));
}

void load_view(int slot)
{
    printf("Load viewpoint #%d\n", slot);
//     global_xf.read(vp_filenames[slot-1]);
    loadparams(savename(base_name, ".p", slot));
    global_xf.read(savename(base_name, ".v", slot));
}

// Keyboard
void specialfunc(int key, int x, int y)
{
    int slot = -1;
    bool save = glutGetModifiers() && GLUT_ACTIVE_SHIFT;
    switch (key) {
        case GLUT_KEY_F1:
            slot = 1;
            break;
        case GLUT_KEY_F2:
            slot = 2;
            break;
        case GLUT_KEY_F3:
            slot = 3;
            break;
        case GLUT_KEY_F4:
            slot = 4;
            break;
        case GLUT_KEY_F5:
            slot = 5;
            break;
        case GLUT_KEY_F6:
            slot = 6;
            break;
        case GLUT_KEY_F7:
            slot = 7;
            break;
        case GLUT_KEY_F8:
            slot = 8;
            break;
        case GLUT_KEY_F9:
            slot = 9;
            break;
        case GLUT_KEY_F10:
            slot = 10;
            break;
        case GLUT_KEY_F11:
            slot = 11;
            break;
        case GLUT_KEY_F12:
            slot = 12;
            break;
        /*default:
            fprintf(stderr, "Unknown key %d\n", key);*/
    }
    if (slot > 0) {
        if (save) save_view(slot);
        else load_view(slot);
    }
    need_redraw();
}

void usage(const char *myname)
{
    fprintf(stderr, "Usage: %s infile...\n", myname);
    exit(1);
}


void hsv2rgb(GLfloat H, GLfloat S, GLfloat V, GLfloat &R, GLfloat &G, GLfloat &B)
{
    //H = H / 360.0;
    if ( S == 0 )                       //HSV from 0 to 1
    {
        R = V;
        G = V;
        B = V;
    }
    else
    {
        GLfloat var_h, var_i, var_r, var_g, var_b, var_1, var_2, var_3;
        var_h = H * 6;
        if ( var_h == 6 ) var_h = 0;      //H must be < 1
        var_i = int( var_h );             //Or ... var_i = floor( var_h )
        var_1 = V * ( 1 - S );
        var_2 = V * ( 1 - S * ( var_h - var_i ) );
        var_3 = V * ( 1 - S * ( 1 - ( var_h - var_i ) ) );

        if      ( var_i == 0 ) { var_r = V     ; var_g = var_3 ; var_b = var_1 ; }
        else if ( var_i == 1 ) { var_r = var_2 ; var_g = V     ; var_b = var_1 ; }
        else if ( var_i == 2 ) { var_r = var_1 ; var_g = V     ; var_b = var_3 ; }
        else if ( var_i == 3 ) { var_r = var_1 ; var_g = var_2 ; var_b = V ;     }
        else if ( var_i == 4 ) { var_r = var_3 ; var_g = var_1 ; var_b = V ;     }
        else                   { var_r = V     ; var_g = var_1 ; var_b = var_2 ; }

        R = var_r; 
        G = var_g;
        B = var_b;
    }
}


void error_message(const char* message)
{
    printf("%s\n", message);
    exit(1);
}


void warning_message(const char* message)
{
    printf("%s\n", message);
}


char* read_file(const char *fn)
{
    FILE *fp;
    char *content = NULL;
    int f, count;
    f = open(fn, O_RDONLY);
    count = lseek(f, 0, SEEK_END);
    close(f);
    if (fn != NULL) {
        fp = fopen(fn,"rt");
        if (fp != NULL) {
            if (count > 0) {
                content = (char *)malloc(sizeof(char) * (count+1));
                count = fread(content,sizeof(char),count,fp);
                content[count] = '\0';
            }
            fclose(fp);
        }
    }
    return content;
}


void printInfoLog(GLhandleARB obj)
{
    int infologLength = 0;
    int charsWritten  = 0;
    char *infoLog;

    glGetObjectParameterivARB(obj, GL_OBJECT_INFO_LOG_LENGTH_ARB,
                                         &infologLength);
    if (infologLength > 0)
    {
        infoLog = (char *)malloc(infologLength);
        glGetInfoLogARB(obj, infologLength, &charsWritten, infoLog);
        printf("%s\n",infoLog);
        free(infoLog);
    }
}


GLhandleARB load_shader(const char* filename, ShaderType type, GLhandleARB program)
{
    GLhandleARB shader;
    if (type == VERTEX_SHADER) 
        shader = glCreateShaderObjectARB(GL_VERTEX_SHADER_ARB);
    else if (type == FRAGMENT_SHADER)
        shader = glCreateShaderObjectARB(GL_FRAGMENT_SHADER_ARB);
    else
        error_message("Unknown shader");
    char *shader_string = read_file(filename);
    const char *str = shader_string;
    glShaderSourceARB(shader, 1, &str, NULL);
    free(shader_string);
    glCompileShaderARB(shader);
    cout << ">>> shader: " << filename << "--------------------" << endl;
    printInfoLog(shader);
    glAttachObjectARB(program, shader);
    return shader;
}



void setup_shaders()
{
    pass1_program = glCreateProgramObjectARB();
    pass2_program = glCreateProgramObjectARB();
    pass3_program = glCreateProgramObjectARB();
    ar_vs = load_shader("./ar.vs", VERTEX_SHADER, pass1_program);
    ar_fs_1p = load_shader("./ar-1p.fs", FRAGMENT_SHADER, pass1_program);
    ar_fs_2p = load_shader("./ar-2p.fs", FRAGMENT_SHADER, pass2_program);
    ar_fs_3p = load_shader("./ar-3p.fs", FRAGMENT_SHADER, pass3_program);
    glLinkProgramARB(pass1_program);
    glLinkProgramARB(pass2_program);
    glLinkProgramARB(pass3_program);

    //tau_location = glGetUniformLocation(pass1_program, "tau");
    //glUniform1f(tau_location, tau);

    // attributes
    //pdir1_location = glGetAttribLocationARB(program, "pdir1");

    k1MAX = std::max(std::abs(k1max), std::abs(k1min));
    GLint uniformLoc = glGetUniformLocation(pass1_program, "k1max");
    if (uniformLoc != -1) {
        cout << k1MAX << endl;
        glUniform1f(uniformLoc, k1MAX);
    }

}

void texture_setup()
{
    glDeleteTextures(1, &textures[0]);
    glDeleteBuffers(1, &depth_rbo);

    // create textures
    glGenTextures(1, textures);

    //glActiveTexture(GL_TEXTURE0);
    glBindTexture(GL_TEXTURE_2D, textures[0]);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA16F, textureWidth, textureHeight, 
                                                       0, GL_RGBA, GL_FLOAT, 0);
    glBindTexture(GL_TEXTURE_2D, 0);

    // create depth renderbuffer
    glGenRenderbuffersEXT(1, &depth_rbo);
    glBindRenderbufferEXT(GL_RENDERBUFFER_EXT, depth_rbo);
    glRenderbufferStorageEXT(GL_RENDERBUFFER_EXT, GL_DEPTH_COMPONENT, 
                             textureWidth, textureHeight);
    glBindRenderbufferEXT(GL_RENDERBUFFER_EXT, 0);

    // create framebuffer
    //glGenFramebuffersEXT(1, &fbo);
    glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, fbo);
    // attach texture to fbo
    glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT, 
                             GL_TEXTURE_2D, textures[0], 0);
    // attach depth rbo to fbo
    glFramebufferRenderbufferEXT(GL_FRAMEBUFFER_EXT, GL_DEPTH_ATTACHMENT_EXT, 
                             GL_RENDERBUFFER_EXT, depth_rbo);
    glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);
    // check
    bool fbo_ok = (glCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT) == GL_FRAMEBUFFER_COMPLETE_EXT);
    cout << "checking FBO: " <<  (fbo_ok ? "ok" : "not ok") << endl;
}

void size_change(int w, int h)
{
    glViewport(0, 0, w, h);
    GLint i, j;
    GLfloat xInc, yInc;

    windowWidth = textureWidth = w;
    windowHeight = textureHeight = h;
    //textureWidth = w/2;
    //textureHeight = h/2;
    cout << "resize(" << w << "x" << h << ")" << endl;

    texture_setup();

    // find largest power-of-two texture smaller than window
    if (!npotTexturesAvailable)
    {
        // Find the largest power of two that will fit in window.

        // Try each width until we get one that's too big
        i = 0;
        while ((1 << i) <= textureWidth)
            i++;
        textureWidth = (1 << (i-1));

        // Now for height
        i = 0;
        while ((1 << i) <= textureHeight)
            i++;
        textureHeight = (1 << (i-1));
    }

    if (textureWidth > maxTexSize) {
        textureWidth = maxTexSize;
    }
    if (textureHeight > maxTexSize) {
        textureHeight = maxTexSize;
    }

    xInc = 1.0f / (GLfloat)textureWidth;
    yInc = 1.0f / (GLfloat)textureHeight;

    for (j = 0; j < 3; j++)
    {
        for (i = 0; i < 3; i++)
        {
            printf("i:%d ", ((j*3)+i)*2 );
            texCoordOffsets[(((j*3)+i)*2)+0] = (-1.0f * xInc) + ((GLfloat)i * xInc);
            texCoordOffsets[(((j*3)+i)*2)+1] = (-1.0f * yInc) + ((GLfloat)j * yInc);
            printf("(%d,%d) = %f,%f\n",i,j,
                texCoordOffsets[(((i*3)+j)*2)+0], texCoordOffsets[(((i*3)+j)*2)+1]);
        }
    }
    
    /*printf("w, h %d,%d\n", w, h);
    printf("windowWidth, windowHeight %d,%d\n", windowWidth, windowHeight);
    printf("textureWidth, textureHeight %d,%d\n", textureWidth, textureHeight);*/
}


void dispose_shaders()
{
    printf("Disposing shaders\n");
    glDeleteObjectARB(ar_vs);
    glDeleteObjectARB(ar_fs_2p);
    glDeleteObjectARB(pass1_program);
    glDeleteObjectARB(pass2_program);
}


// Compute a "feature size" for the mesh: computed as 1% of
// the reciprocal of the 10-th percentile curvature
void compute_feature_size(TriMesh* themesh)
{
	int nv = themesh->curv1.size();
	int nsamp = min(nv, 500);

	vector<float> samples;
	samples.reserve(nsamp * 2);

	for (int i = 0; i < nsamp; i++) {
		// Quick 'n dirty portable random number generator
		static unsigned randq = 0;
		randq = unsigned(1664525) * randq + unsigned(1013904223);

		int ind = randq % nv;
		samples.push_back(fabs(themesh->curv1[ind]));
		samples.push_back(fabs(themesh->curv2[ind]));
	}

	const float frac = 0.1f;
	const float mult = 0.01f;
	themesh->need_bsphere();
	float max_feature_size = 0.05f * themesh->bsphere.r;

	int which = int(frac * samples.size());
	nth_element(samples.begin(), samples.begin() + which, samples.end());

	feature_size = min(mult / samples[which], max_feature_size);
}

// Diffuse the normals across the mesh
void filter_normals(TriMesh* themesh, float smooth)
{
	diffuse_normals(themesh, smooth);
	themesh->curv1.clear();
	themesh->dcurv.clear();
	themesh->need_curvatures();
	themesh->need_dcurv();
}


// Diffuse the curvatures across the mesh
void filter_curv(TriMesh* themesh, float smooth)
{
	diffuse_curv(themesh, smooth);
	themesh->dcurv.clear();
	themesh->need_dcurv();
}



int main(int argc, char *argv[])
{
    if (argc < 2)
        usage(argv[0]);
        
    char* arsize = getenv("AR_SIZE");
    if (arsize) {
        cout << "AR_SIZE:" << arsize << endl;
        windowWidth = windowHeight = atoi(arsize);
    }
    
    // startup
    glutInitWindowSize(windowWidth, windowHeight);
    glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGBA | GLUT_DEPTH);
    glutInit(&argc, argv);

    // mesh setup
    const char *filename = argv[1];
    mesh = TriMesh::read(filename);
    if (!mesh)
        usage(argv[0]);
    mesh->need_bsphere();
    float radius = mesh->bsphere.r;
    //scale(mesh, 1.0f/radius);
    //mesh->bsphere.valid = false;
    //printf(">>> radius: %f\n", mesh->bsphere.r);
    
    mesh->need_tstrips();
    mesh->need_bsphere();
    mesh->need_normals();
    mesh->need_curvatures();
    mesh->need_dcurv();
    mesh->need_adjacentfaces();
    
    compute_feature_size(mesh);
    //float smooth = 0.5f * mesh->feature_size();
    //filter_curv(mesh, 1.1*smooth);
    //filter_normals(mesh, 2*smooth);

    feature_size = mesh->feature_size();
    printf(">>> Feature size: %f\n", feature_size);

    //edgeflip(mesh);
    //diffuse_normals(mesh, 0.4);
    //tau = 1.0;
    // feat. size x ar_thresg linear relation (experimentally determined)
    //tau = 4.24158132*log10(feature_size)-5.19257055;

    int nv = mesh->vertices.size();
    ndotv.resize(nv);
    //curvatures.resize(nv);
    q1.resize(nv);
    t1.resize(nv);
    Dt1q1.resize(nv);

    curv = new GLfloat[3*nv];
    curvatures = new GLfloat[3*nv];
    /*k1max = std::fabs(mesh->curv1[0]);
    k2max = std::fabs(mesh->curv2[0]);
    k1min = std::fabs(mesh->curv1[0]);
    k2min = std::fabs(mesh->curv2[0]);*/
    k1max = mesh->curv1[0];
    k2max = mesh->curv2[0];
    k1min = mesh->curv1[0];
    k2min = mesh->curv2[0];

    for (int j = 0; j < nv; j++) {
        /*if (std::fabs(mesh->curv1[j]) > k1max) k1max = std::fabs(mesh->curv1[j]);
        if (std::fabs(mesh->curv2[j]) > k2max) k2max = std::fabs(mesh->curv2[j]);
        if (std::fabs(mesh->curv1[j]) < k1min) k1min = std::fabs(mesh->curv1[j]);
        if (std::fabs(mesh->curv2[j]) < k2min) k2min = std::fabs(mesh->curv2[j]);*/
        if (mesh->curv1[j] > k1max) k1max = mesh->curv1[j];
        if (mesh->curv2[j] > k2max) k2max = mesh->curv2[j];
        if (mesh->curv1[j] < k1min) k1min = mesh->curv1[j];
        if (mesh->curv2[j] < k2min) k2min = mesh->curv2[j];
        curvatures[3*j+0] = mesh->curv1[j];
        curvatures[3*j+1] = mesh->curv2[j];
        curvatures[3*j+2] = 0.0;
        k1mean += mesh->curv1[j];
        k2mean += mesh->curv2[j];
    }
    k1mean /= (double)nv;
    k2mean /= (double)nv;
    for (int j = 0; j < nv; j++) {
        k1var += 2.0*abs(k1mean - mesh->curv1[j]);
        k2var += 2.0*abs(k2mean - mesh->curv2[j]);
    }
    k1var /= (double)nv;
    k2var /= (double)nv;
    printf("k1max: %f, k2max: %f\n", k1max, k2max);
    printf("k1min: %f, k2min: %f\n", k1min, k2min);
    printf("k1mean: %f, k2mean: %f\n", k1mean, k2mean);
    printf("k1var: %f, k2var: %f\n", k1var, k2var);
    
    //mesh->need_faces();
    /*cout << "Faces: " << mesh->faces.size() << endl;
    for(int i=0; i<mesh->faces.size(); i++) {
        int* v = mesh->faces[i].v;
        float v01 = mesh->pdir1[v[0]] DOT mesh->pdir1[v[1]];
        if (v01 < 0.0) mesh->pdir1[v[1]] = -mesh->pdir1[v[1]];
        float v12 = mesh->pdir1[v[1]] DOT mesh->pdir1[v[2]];
        if (v12 < 0.0) mesh->pdir1[v[2]] = -mesh->pdir1[v[2]];

        v01 = mesh->pdir2[v[0]] DOT mesh->pdir2[v[1]];
        if (v01 < 0.0) mesh->pdir2[v[1]] = -mesh->pdir2[v[1]];
        v12 = mesh->pdir2[v[1]] DOT mesh->pdir2[v[2]];
        if (v12 < 0.0) mesh->pdir2[v[2]] = -mesh->pdir2[v[2]];
    }*/
    /*for(int i=0; i<mesh->faces.size(); i++) {
        int* v = mesh->faces[i].v;
        float v01 = mesh->pdir1[v[0]] DOT mesh->pdir1[v[1]];
        float v12 = mesh->pdir1[v[1]] DOT mesh->pdir1[v[2]];
        float v20 = mesh->pdir1[v[2]] DOT mesh->pdir1[v[0]];
        if (v01 < 0.0 || v12 < 0.0 || v20 < 0.0) cout << "blih!" << endl;
    }*/
   
    //mesh->need_tstrips();    
    //mesh->need_bsphere();
    //tau = 2.6542247861*log(mesh->bsphere.r)-12.8179214442;
    //tau = 2.5991676251*log(mesh->bsphere.r)-11.7410160716;
    //tau = 1.8*log(mesh->bsphere.r)-8.5;

    
    //scale(mesh, 1.0/r);
    compute_feature_size(mesh);
    feature_size = mesh->feature_size();
    printf(">>> Feature size: %f\n", feature_size);
    
    tau = sqr(feature_size);
    cout << "tau: " << tau << endl;
    cout << "lambda: " << lambda << endl;

    xffilename = xfname(filename);
    base_name = getbase_name(filename);
    printf("modelname: %s\n", savename(base_name, ".v", 7).c_str());
    printf("modelname: %s\n", savename(base_name, ".p", 11).c_str());
    xf = xform();
    modelname = get_model_name(base_name);
    
    
    // ---------------------------------------------------
    glutCreateWindow("Apparent Ridges");
    glutReshapeFunc(size_change);
    glutDisplayFunc(redraw);
    glutMouseFunc(mousebuttonfunc);
    glutMotionFunc(mousemotionfunc);
    glutKeyboardFunc(keyboardfunc);
    glutSpecialFunc(specialfunc);
    glutIdleFunc(idle);
    
    // setup
    //if (GLEW_VERSION_2_0 || GLEW_ARB_texture_non_power_of_two) {
    if (GL_VERSION_2_0 || GL_ARB_texture_non_power_of_two) {
        npotTexturesAvailable = GL_TRUE;
    }
    else {
        warning_message("No support for non power-of-two texture\n");
    }

    glGetIntegerv(GL_MAX_TEXTURE_SIZE, &maxTexSize);
    printf("Maximum texture size: %d\n", maxTexSize);

    resetview();

    glewInit();
    if (!GLEW_ARB_vertex_shader || !GLEW_ARB_fragment_shader) {
        error_message("Sorry. No support for GLSL\n");
    }

    #ifdef FBO
    // create textures
    glGenTextures(2, textures);

    //glActiveTexture(GL_TEXTURE0);
    glBindTexture(GL_TEXTURE_2D, textures[0]);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
    //glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA8, textureWidth, textureHeight, 0, GL_RGBA, GL_UNSIGNED_BYTE, 0);
    //glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA16, textureWidth, textureHeight, 0, GL_RGBA, GL_UNSIGNED_INT, 0);
    //glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA16F, textureWidth, textureHeight, 0, GL_RGBA, GL_FLOAT, 0);
    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA32F, textureWidth, textureHeight, 0, GL_RGBA, GL_FLOAT, 0);

    //glActiveTexture(GL_TEXTURE1);
    glBindTexture(GL_TEXTURE_2D, textures[1]);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
    //glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, textureWidth, textureHeight, 0, GL_RGBA, GL_FLOAT, 0);
    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA32F, textureWidth, textureHeight, 0, GL_RGBA, GL_FLOAT, 0);

    /*//glActiveTexture(GL_TEXTURE2);
    glBindTexture(GL_TEXTURE_2D, textures[2]);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE); // ***
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    //glTexParameteri(GL_TEXTURE_2D, GL_GENERATE_MIPMAP, GL_TRUE);
    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, textureWidth, textureHeight, 0, GL_RGBA, GL_FLOAT, 0);
    */

    glBindTexture(GL_TEXTURE_2D, 0);

    // create depth renderbuffer
    glGenRenderbuffersEXT(1, &depth_rbo);
    glBindRenderbufferEXT(GL_RENDERBUFFER_EXT, depth_rbo);
    glRenderbufferStorageEXT(GL_RENDERBUFFER_EXT, GL_DEPTH_COMPONENT, 
                             textureWidth, textureHeight);
    glBindRenderbufferEXT(GL_RENDERBUFFER_EXT, 0);

    // create framebuffer
    glGenFramebuffersEXT(1, &fbo);
    glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, fbo);
    // attach texture to fbo
    glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT, 
                             GL_TEXTURE_2D, textures[0], 0);
    // attach depth rbo to fbo
    glFramebufferRenderbufferEXT(GL_FRAMEBUFFER_EXT, GL_DEPTH_ATTACHMENT_EXT, 
                             GL_RENDERBUFFER_EXT, depth_rbo);
    glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);
    // check
    bool fbo_ok = (glCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT) == GL_FRAMEBUFFER_COMPLETE_EXT);
    cout << "checking FBO: " <<  (fbo_ok ? "ok" : "not ok") << endl;
    #endif


    #ifdef VBO
    cout << "VBO enabled" << nv << endl;
    // VBO setup
    glGenBuffersARB(1, &vertex_vbo);
    glGenBuffersARB(1, &normal_vbo);
    glGenBuffersARB(1, &curv_vbo);
    glGenBuffersARB(1, &pdir_vbo);
    glGenBuffersARB(1, &index_vbo);
    glBindBufferARB(GL_ARRAY_BUFFER_ARB, vertex_vbo);
    glBufferDataARB(GL_ARRAY_BUFFER_ARB, sizeof(mesh->vertices[0])*nv, &mesh->vertices[0][0], GL_STATIC_DRAW_ARB);
    glBindBufferARB(GL_ARRAY_BUFFER_ARB, normal_vbo);
    glBufferDataARB(GL_ARRAY_BUFFER_ARB, sizeof(mesh->normals[0])*nv, &mesh->normals[0][0], GL_STATIC_DRAW_ARB);
    glBindBufferARB(GL_ARRAY_BUFFER_ARB, curv_vbo);
    glBufferDataARB(GL_ARRAY_BUFFER_ARB, sizeof(GLfloat)*3*nv, curvatures, GL_STATIC_DRAW_ARB);
    glBindBufferARB(GL_ARRAY_BUFFER_ARB, pdir_vbo);
    glBufferDataARB(GL_ARRAY_BUFFER_ARB, sizeof(mesh->pdir1[0])*3*nv, &mesh->pdir1[0][0], GL_STATIC_DRAW_ARB);

    glBindBufferARB(GL_ELEMENT_ARRAY_BUFFER_ARB, index_vbo);
    glBufferDataARB(GL_ELEMENT_ARRAY_BUFFER_ARB, sizeof(GLuint)*mesh->tstrips.size(), &mesh->tstrips[0], GL_STATIC_DRAW_ARB);

    glBindBufferARB(GL_ARRAY_BUFFER_ARB, 0);
    glBindBufferARB(GL_ELEMENT_ARRAY_BUFFER_ARB, 0);
    #endif

    // setup shader programs
    setup_shaders();
    
    // prepare display list
    dlist = glGenLists(1);
    //if(!dlist) cerr << "Error: display list not available" << endl;
    prepare_list();



    // ----------------- main loop --------------------
    glutMainLoop();

    // finalization   
    dispose_shaders();
    glDeleteLists(dlist, 1);
    #ifdef FBO
    glDeleteTextures(1, textures);
    glDeleteBuffers(1, &depth_rbo);
    #endif
    #ifdef VBO
    glDeleteBuffers(1, &vertex_vbo);
    glDeleteBuffers(1, &normal_vbo);
    glDeleteBuffers(1, &curv_vbo);
    glDeleteBuffers(1, &pdir_vbo);
    glDeleteBuffers(1, &index_vbo);
    #endif
}


